# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from test_util import CommandLineT


class ExamplesTests(CommandLineT):
    def test_simple(self):
        """ Test that examples/simple"""

        # retrieve the configuration and data files
        self.run("cp {0}/examples/simple/gauss2D.data .".format(self.ask_path))
        self.run("cp {0}/examples/simple/simple.conf .".format(self.ask_path))

        r = self.run("ask simple.conf")
        assert "output/labelled.data" in r.files_created,\
               "labelled.data is created"
        assert "output/plot00000.png" in r.files_created,\
               "a plot is created"

    def test_gauss2D(self):
        """ Test that examples/gauss2D"""

        # retrieve the configuration and data files
        self.run("cp {0}/examples/gauss2D/source.R ."
                .format(self.ask_path))

        # consider only 300 points in the test set to make the test faster
        data =[]
        f = file("{0}/examples/gauss2D/test.data".format(self.ask_path))
        for l in range(300):
            data.append(f.readline())
        f.close()
        self.writefile("test.data", "".join(data))

        # change the default iterations values to make the test faster
        import json
        original_conf = json.loads(file("{0}/examples/gauss2D/experiment.conf"
                                   .format(self.ask_path)).read())
        original_conf["modules"]["control"]["params"]["n"] = 150
        self.conf(original_conf)

        r = self.run("ask test.conf")
        assert "output/labelled.data" in r.files_created,\
               "labelled.data is created"
        assert "output/plot00002.png" in r.files_created,\
               "a plot is created"



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import numpy as np
import sys
import os
import subprocess

def fatal(msg, exitcode = 1):
    """
    Reports fatal error and die with exitcode.
    """
    sys.stderr.write(str(msg) + "\n")
    sys.exit(exitcode)

def find_closest(points, p):
    return min(zip(abs(points-p),points))[1]

def multiselect(x, set):
    """
    Returns a boolean np.array that contains True for the elements
    in x that are in set.
    """
    if len(set) == 0:
        return np.array([False]*x.size)
    l = x == set[0]
    for v in set[1:]:
        l = l | (x == v)
    return l

def get_cov_ub(mean, sd, card, confidence):
    """
    Returns the confidence interval upper-bound of the coefficient of variance.

    mean: empirical mean of the distribution
    sd: empirical standard deviation of the distribution
    card: number of samples in the distribution
    confidence: 0.9 means the upper-bound is correct 90% of the time
    """
    coef_path = os.path.join(os.environ["ASKHOME"],"./common/coef-variation.R")
    p = subprocess.Popen(map(str,[coef_path, mean, sd, card, confidence]), 
            stdout=subprocess.PIPE)
    out, err = p.communicate()
    print float(out.split()[1])
    return float(out.split()[1])





# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from copy import deepcopy

from common.tree import Leaf


class OrdinalAxis:
    """
    OrdinalAxis class, tracks range constraints on an ordinal factor
    It is used during a CART tree traversal to accumulate each split node
    constraints. When the traversal is done, and a leaf is reached,
    the get_range method returns the precise range associated with the
    reached leaf.
    """
    def __init__(self, minv=-2 ** 32, maxv=2 ** 32):
        """
        Initialize an OrdinalAxis class.

        minv: minimal value of the axis
        maxv: maximal value of the axis
        """
        self.minv = minv
        self.maxv = maxv

    def add_left(self, cut):
        """
        Add a constraint when following a CART splitting node to the left.
        """
        self.maxv = min(cut, self.maxv)

    def add_right(self, cut):
        """
        Add a constraint when following a CART splitting node to the right.
        """
        self.minv = max(cut, self.minv)

    def get_range(self):
        """
        Retrieve the range that satisfies all the left and right constraints
        added.
        """
        return {"min": self.minv, "max": self.maxv}


class CategoricalAxis:
    """
    OrdinalAxis class, tracks range constraints on an categorical factor
    It is used during a CART tree traversal to accumulate each split node
    constraints. When the traversal is done, and a leaf is reached,
    the get_range method returns the precise range associated with the
    reached leaf.
    """
    def __init__(self, tags):
        """
        Initialize an CategoricalAxis class.

        tags: possible values of the categorical axis
        """

        self.belongs_to = set(range(len(tags)))
        self.tags = tags

    def add_left(self, cut):
        """
        Add a constraint when following a CART splitting node to the left.
        """
        self.add(cut[0])

    def add_right(self, cut):
        """
        Add a constraint when following a CART splitting node to the right.
        """
        self.add(cut[1])

    def add(self, ok):
        """
        Register the ''ok'' constraints, generic function used by add_left
        and add_right.
        """
        self.belongs_to = self.belongs_to.intersection(ok)

    def get_range(self):
        """
        Retrieve the range that satisfies all the left and right constraints
        added.
        """
        out = []
        for i in self.belongs_to:
            out.append(self.tags[i])
        return out


class Constraints:
    """
    Abstract Constraints class.
    It is used during a CART tree traversal to accumulate each split node
    constraints. When the traversal is done, and a leaf is reached
    the get_range method returns the precise range associated with the
    reached leaf.
    """

    def __init__(self, factors):
        """
        Initialize a Constraints class.

        factors: factors tree.py representation
        """
        self.factors = factors
        self.axes = []
        self.constraints = []

        for f in factors:
            self.axes.append(f["name"])
            if f["type"] != "categorical":
                self.constraints.append(
                OrdinalAxis(minv=f["range"]["min"],
                            maxv=f["range"]["max"]))
            else:
                self.constraints.append(
                    CategoricalAxis(f["values"]))

    def add_left(self, axis, cut):
        self.constraints[axis].add_left(cut)

    def add_right(self, axis, cut):
        self.constraints[axis].add_right(cut)

    def get_range(self):
        return dict([(a, c.get_range())
                     for a, c in zip(self.axes, self.constraints)])


def decorate_leaves(tree, C):
    """
    Use a Constraint class to accumulate the splitting constraints
    during a full tree traversal. Decorate all the leaves with the
    associated ranges.

    C: a constraint class
    tree: a tree
    """
    if isinstance(tree, Leaf):
        tree.constraint = C
        yield tree
    else:
        #left
        CL = deepcopy(C)
        CL.add_left(tree.axis, tree.cut)
        for n in decorate_leaves(tree.left, CL):
            yield n
        #right
        CR = deepcopy(C)
        CR.add_right(tree.axis, tree.cut)
        for n in decorate_leaves(tree.right, CR):
            yield n


def get_tagged_models(factors, T):
    """
    Returns an array containing each leaf in the tree 'T',
    its associated model,
    its associated range constraint,
    and a pointer to the leaf itself.

    factors: factors tree.py representation
    T: a tree
    """
    C = Constraints(factors)
    leaves = decorate_leaves(T, C)
    return [{"model":list(l.model),
            "error": l.error,
            "constraint": l.constraint.get_range(),
            "leaf": l}
            for l in leaves]



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

def residues(d,T):
    """ 
    Computes the residues between a CART prediction tree ''T'' and a 
    set of data ''d''.
    """
    x = d[:,:-1]

    y = map(lambda x: T.compute(x), x)
    err = []
    for a,b in zip(d[:,-1],y):
        err.append((a-b))
    return err



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import pickle
import tempfile

from colors import colors
from regression import linear


def tree2dot(outfilename, tree, axes, categories):
    """
    Dumps a tree in dotty format.

    outfilename : name of the output dot file
    tree : the tree to dump
    axes : a list of strings containing the label for the axes
    categories : a list of dictionnaries that returns for each
                 categorical axis, the labels of each category.
                 For instance, if dimension 3 contains three
                 classes: 1-> blue, 2-> red, 3 -> green
                 then categories[3] = {1:"blue", 2:"red", 3:"green"}
    """
    f = open(outfilename, "w")
    f.write("graph decisionT {\n" + tree.todot(axes, categories) + "\n}")
    f.close()


def tree2png(outfilename, tree, axes, categories):
    """
    Dumps a tree in png.

    outfilename : name of the output png file
    tree : the tree to dump
    axes : a list of strings containing the label for the axes
    categories : a list of dictionnaries that returns for each
                 categorical axis, the labels of each category.
                 For instance, if dimension 3 contains three
                 classes: 1-> blue, 2-> red, 3 -> green
                 then categories[3] = {1:"blue", 2:"red", 3:"green"}
    """
    dotfile = tempfile.NamedTemporaryFile()
    tree2dot(dotfile.name, tree, axes, categories)
    os.system("dot -Tpng -o {0} {1}".format(outfilename, dotfile.name))
    dotfile.close()


def Lformat(L, pl=6):
    """
    Line format utility function.
    L: a list of tags.
    pl: the number of tags per line

    returns a string representation of the tags with exactly pl tags
    per line.
    """
    lines = []
    for i, v in enumerate(L):
        lines.append(v)
        if i % pl == pl - 1:
            lines.append("\\n")
    return ",".join(lines)


def tag_leaves(tree):
    """
    Decorates the leaves of a tree with consecutive numeric tags
    going from 0 to NUMBER_OF_LEAVES-1.

    The tag is added inside each leaf object in parameter tag
    (leaf.tag).

    Returns NUMBER_OF_LEAVES-1.

    As an example consider this three leaved tree:
    >>> T = Node(Node(Leaf([0]),\
                      Leaf([1]),\
                      0,0,[]),\
                 Leaf([2]),\
                 0,0,[])
    >>> tag_leaves(T)
    2
    >>> list(map(lambda l: l.tag, leaf_iterator(T)))
    [0, 1, 2]
    """
    leaves = list(leaf_iterator(tree))
    for i, leaf in enumerate(leaves):
        leaf.tag = i
    return len(leaves) - 1


def leaf_iterator(tree):
    """
    Returns a generator that iterates over all leaves of a tree
    in depth first order.

    As an example consider this three leaved tree:
    >>> T = Node(Node(Leaf([0]),\
                      Leaf([1]),\
                      0,0,[]),\
                 Leaf([2]),\
                 0,0,[])
    >>> list(leaf_iterator(T))
    [L([0]), L([1]), L([2])]
    """
    if isinstance(tree, Leaf):
        yield tree
    else:
        for subtree in [tree.left, tree.right]:
            for n in leaf_iterator(subtree):
                yield n


def node_iterator(tree):
    """
    Returns a generator that iterates over all nodes of a tree
    in depth first order.

    """
    if isinstance(tree, Leaf):
        yield tree
    else:
        yield tree
        for subtree in [tree.left, tree.right]:
            for n in node_iterator(subtree):
                yield n


def save_tree(T, output_file):
    """
    Save a tree T to a file.
    """
    f = open(output_file, "w")
    pickle.dump(T, f)
    f.close()


def load_tree(input_file):
    """
    Load a tree T from a file.
    """
    f = open(input_file, "r")
    l = pickle.load(f)
    f.close()
    return l


class Node():
    """
    Regression tree internal node.
    """
    def __init__(self, left, right, axis, cut, model, categorical=False):
        """
        left, right (Node) : left and right subtrees
        axis (int) : axis over which this node splits the domain
        cut : cut point for this node
              * for ordinal data: a value v, the cut is done for
                (all x < v , all x >= v)
              * for categorical data: a 2-uplet containing the
                two classes. For example to separate odd and even
                categories ([1,3,5], [2,4,6]).
        model : the linear model associated to this subtree
        categorical : is this cut categorical ?
        """
        self.left = left
        self.right = right
        self.axis = axis
        self.model = model
        self.cut = cut
        self.categorical = categorical
        self.data = []

    def fill(self, point):
        """
        point: array of coordinates for a point
        returns the value predicted by the tree for the
        given point.
        """
        self.data.append(point)
        p = point[self.axis]
        if self.categorical:
            is_left = (p in self.cut[0])
        else:
            is_left = p < self.cut

        if is_left:
            return self.left.fill(point)
        else:
            return self.right.fill(point)

    def compute(self, point):
        """
        point: array of coordinates for a point
        returns the value predicted by the tree for the
        given point.
        """
        p = point[self.axis]
        if self.categorical:
            is_left = (p in self.cut[0])
        else:
            is_left = p < self.cut

        if is_left:
            return self.left.compute(point)
        else:
            return self.right.compute(point)

    def whichnode(self, point):
        """
        point: array of coordinates for a point
        returns the leaf responsible for modeling
        the passed point.
        """
        p = point[self.axis]
        if self.categorical:
            is_left = p in self.cut[0]
        else:
            is_left = p < self.cut

        if is_left:
            return self.left.whichnode(point)
        else:
            return self.right.whichnode(point)

    def whichmodel(self, point):
        """
        point: array of coordinates for a point
        returns the tag of the leaf responsible for modeling
        the passed point.
        """
        p = point[self.axis]
        if self.categorical:
            is_left = p in self.cut[0]
        else:
            is_left = p < self.cut

        if is_left:
            return self.left.whichmodel(point)
        else:
            return self.right.whichmodel(point)

    def todot(self, axes=None, categories=None):
        """
        Internal function to output the tree in dot format.
        """
        out = []
        if self.categorical:
            labelsl = Lformat([categories[self.axis][i] for i in self.cut[0]])
            labelsr = Lformat([categories[self.axis][i] for i in self.cut[1]])
            out.append("{0} [label=\"{1}\"];"
                       .format(id(self), axes[self.axis]))
        else:
            labelsr = labelsl = ""
            out.append("{0} [label=\"{1} < {2}\"];"
                       .format(id(self), axes[self.axis], "%.2f" % self.cut))

        out.append(self.left.todot(axes, categories))
        out.append(self.right.todot(axes, categories))

        out.append("{0} -- {1} [label=\"{2}\"];"
                   .format(id(self), id(self.left), labelsl))
        out.append("{0} -- {1} [label=\"{2}\"];"
                   .format(id(self), id(self.right), labelsr))
        return "\n".join(out)


class Leaf(Node):
    """
    Regression tree leaf node.
    """
    def __init__(self, model, error=None):
        """
        model : the linear model associated to this leaf
        """
        self.model = model
        self.error = error
        self.data = []
        self.future_points = 0
        self.tag = -1

    def compute(self, point):
        """
        This function computes the prediction of Tree ''self'' on
        point ''point''.
        """
        regression_d = len(self.model) - 1
        if regression_d == 0:
            return self.model[0]
        else:
            p = point[-regression_d:]
            return linear(self.model, *p)

    def fill(self, point):
        self.data.append(point)

    def whichmodel(self, point=None):
        return self.tag

    def whichnode(self, point=None):
        return self

    def __repr__(self):
        return "L({0})".format(repr(self.model))

    def todot(self, axes=None, categories=None):
        model = "*%s*|" % self.tag + "|".join(["%.4f" % v for v in self.model])
        return (
            "{0} [label=\"{{{1}}}\", shape=\"record\",color=\"{2}\"];"
                .format(id(self), model, colors[self.tag % len(colors)]))

if __name__ == "__main__":
    import doctest
    doctest.testmod()



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from test_util import CommandLineT


class BootstrapFileSource(CommandLineT):
    def test_missing_params(self):
        """ File source should print a message if parameters are missing """
        self.conf({"modules": {"source": {"params": {}}}})

        r = self.run_module("source/file",
                            "test.conf request sample",
                            expect_error=True)
        assert r.stderr.find("data_file") != -1,\
                 "error message should mention data_file"

    def test_request(self):
        """ Source request should be fullfilled """
        card = 3
        self.conf(
            {"modules": {"source":
                         {"params": {"data_file": self.tfile("test1.data")}}}})

        self.writefile("request", "\n".join(["5 3000", "3 2777", "8 2777"]))
        r = self.run_module("source/file",
                            "test.conf request sample",
                            expect_stderr=True)

        assert "sample" in r.files_created, "sample is created"
        sample = r.files_created["sample"]
        lines = [l for l in sample.bytes.split("\n") if l]
        sample_card = len(lines)
        assert sample_card == card, "we should get exactly 3 samples"
        for l in lines:
            print l
            assert len(l.split()) == 3, "each line should have 3 elements)"

    def test_missing(self):
        """ Missing points should not be returned """
        card = 2
        self.conf(
            {"modules": {"source":
                         {"params": {"data_file": self.tfile("test1.data")}}}})

        self.writefile("request",
                       "\n".join(["5 3000", "3 2777", "100000000 2777"]))
        r = self.run_module("source/file",
                            "test.conf request sample",
                            expect_stderr=True)
        sample = r.files_created["sample"]
        lines = [l for l in sample.bytes.split("\n") if l]
        sample_card = len(lines)
        assert sample_card == card, "we should get exactly 2 samples"



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from test_util import CommandLineT


class AmartTests(CommandLineT):
    def test_basic_amart(self):
        """ Test amart with ask """
        self.conf({
            "output_directory": "outdir",

            "factors": [
            {"name": "k",
             "type": "integer",
             "range": {"min": 1, "max": 8}},
            {"name": "n",
             "type": "integer",
             "range": {"min": 5, "max": 4999}}],

            "modules": {
            "sampler": {
            "executable": "sampler/amart",
            "params": {"n": 50, "seeds": 3, "trees": 10}},

            "control": {
            "executable": "control/points",
            "params": {"n": 150}},

            "bootstrap": {
            "executable": "bootstrap/random-file",
            "params": {"data_file": self.tfile("test1.data"), "n": 50}},

            "source": {
            "executable": "source/file",
            "params": {"data_file": self.tfile("test1.data")}}}})

        r = self.run("ask test.conf")
        assert "outdir/labelled.data" in r.files_created,\
               "labelled.data is created"
        lines = r.files_created["outdir/labelled.data"].bytes.split("\n")
        card = len([l for l in lines if l])
        assert card == 150, "150 samples should have been obtained"


class HierarchicalTests(CommandLineT):
    def test_basic_hierarchical(self):
        """ Test hierarchical with ask """
        self.conf({
            "output_directory": "outdir",
            "factors": [
            {"name": "k",
             "type": "integer",
             "range": {"min": 1, "max": 8}},
            {"name": "n",
             "type": "integer",
             "range": {"min": 5, "max": 4999}}],

            "modules": {
            "sampler": {
            "executable": "sampler/hierarchical",
            "params": {"n": 50, "cp": 0.01}},

            "control": {
            "executable": "control/points",
            "params": {"n": 150}},

            "bootstrap": {
            "executable": "bootstrap/random-file",
            "params": {"data_file": self.tfile("test1.data"), "n": 50}},

            "source": {
            "executable": "source/file",
            "params": {"data_file": self.tfile("test1.data")}}}})

        r = self.run("ask test.conf")
        assert "outdir/labelled.data" in r.files_created,\
                 "labelled.data is created"
        lines = r.files_created["outdir/labelled.data"].bytes.split("\n")
        card = len([l for l in lines if l])
        assert card == 150, "150 samples should have beed obtained"



import sys

def string(string):
    try:
        if sys.version_info >= (3,):
            if isinstance(string, str):
                return string
            return str(string, "utf-8")
        else:
	    return string.decode('utf-8')
    except:
        return string



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import unittest
from scripttest import TestFileEnvironment
import json


class CommandLineT(unittest.TestCase):
    def setUp(self):
        self.test_path = os.path.dirname(os.path.realpath(__file__))
        self.ask_path = os.path.realpath(os.path.join(self.test_path, ".."))
        os.environ["ASKHOME"] = self.ask_path
        if "PYTHONPATH" in os.environ:
            os.environ["PYTHONPATH"] += ":" + self.ask_path
        else:
            os.environ["PYTHONPATH"] = self.ask_path
        os.environ["PATH"] = self.ask_path + ":" + os.environ["PATH"]

        # ensure test sandbox exists
        try:
            os.makedirs("sandbox")
        except OSError as exc:
            if exc.errno == os.errno.EEXIST:
                pass
            else:
                raise

        self.E = TestFileEnvironment(os.path.join("sandbox", str(self.id())))
        self.run = self.E.run
        self.writefile = self.E.writefile
        self.clear = self.E.clear

    def run_module(self, module, args, **kwargs):
        """ Run an ask modules by itself """
        m = os.path.join(self.ask_path, module)
        print m
        return self.run(m + " " + args, **kwargs)

    def tfile(self, name):
        """ Shortcut to refer to test files """
        return os.path.join(self.test_path, name)

    def conf(self, data, file_name="test.conf"):
        """ Shortcut to write custom test configurations """
        self.writefile(file_name,
                       json.dumps(data)+"\n")



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Sample colors used in the tree debugging functions

colors = ["Red",
          "Blue",
          "Gray",
          "Yellow",
          "Purple",
          "Green",
          "Maroon",
          "Navy",
          "Orange",
          "Pink",
          "Black"]



#!/usr/bin/env python
# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys

# Check that all the python dependencies are satisfied

missing = []
ubuntu = []

try:
    import numpy
except ImportError:
    missing.append("numpy")
    ubuntu.append("python-numpy")

try:
    import argparse
except ImportError:
    missing.append("argparse")
    ubuntu.append("python-argparse")

try:
    import scipy
except ImportError:
    missing.append("scipy")
    ubuntu.append("python-scipy")


if missing:
    print ("Please install the following python packages: {0}"
        .format("".join(missing)))
    print ("On ubuntu you can type: sudo apt-get install {0}"
        .format(" ".join(ubuntu)))
    sys.exit(1)



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import tempfile
from tree import *


def parse_cart_tree(lines, categories):
    """
    Parses CART trees generated by R module 'rpart'.

    lines: an array of string containing the output of the 'rpart' module
    categories: array that maps each factor to an the categorical values
    accepted by that factor.

    Returns the CART tree in tree.py format.
    """
    nodes = {}
    for l in lines[5:]:
        # read a node
        sp = l.split()
        number = int(sp[0][:-1])

        # Identify the root node
        if sp[1] == "root":
            node = {"prediction": float(sp[4])}
        else:
            # Detect the branch split condition
            if ">" in sp[1] or "<" in sp[1]:
                # Here we treat ordinal factors that are marked with  < or >=

                # detect the direction of the split condition
                # indeed the leaf that satisfies x >= split is not always the
                # right one.
                if ">" in sp[1]:
                    axis = sp[1].split(">=")
                    direction = 1
                else:
                    axis = sp[1].split("<")[0], sp[2]
                    direction = 0
                    sp.pop(2)

                node = {"axis": int(axis[0][1:]) - 1,
                        "cut": float(axis[1]),
                        "prediction": float(sp[4]),
                        "direction": direction}
            else:
                # Here we handle categorial factors that are marked with =
                axis = sp[1].split("=")
                node = {"axis": int(axis[0][1:]) - 1,
                        "choices": map(int, axis[1].split(",")),
                        "prediction": float(sp[4])}

        # Store the parsed node in the nodes table
        nodes[number] = node

    def get_tree(n):
        """
        This internal function returns the n-th tree in the 'rpart'
        representation. The tree is returned in our python tree format defined
        in tree.py. Rpart encodes the root as 1, the left child of n as 2*n,
        and the right child of n as 2*n+1.
        """
        #is it a leaf
        if 2 * n not in nodes:
            return Leaf(model=[nodes[n]["prediction"]])

        #get left and right children
        left = get_tree(2 * n)
        right = get_tree(2 * n + 1)

        if "choices" in nodes[2 * n]:
            lc = nodes[2 * n]["choices"]
            rc = nodes[2 * n + 1]["choices"]
            axis = nodes[2 * n]["axis"]

            chosen = set(lc + rc)
            allp = set(list(range(len(categories[axis]))))
            missing = allp - chosen
            rc += list(missing)

            # Build tree.py categorical Node
            return Node(left, right,
                        axis=axis,
                        cut=(lc,
                             rc),
                        model=[nodes[n]["prediction"]],
                        categorical=True)
        else:
            # Build tree.py ordinal Node
            # (the if is needed to normalize the direction, indeed in
            #  the tree.py representation the leaf satisfying x >= split is
            #  always the right one).
            if nodes[2 * n]["direction"] == 0:
                return Node(left, right,
                            axis=nodes[2 * n]["axis"],
                            cut=nodes[2 * n]["cut"],
                            model=[nodes[n]["prediction"]],
                            categorical=False)
            else:
                return Node(right, left,
                            axis=nodes[2 * n]["axis"],
                            cut=nodes[2 * n]["cut"],
                            model=[nodes[n]["prediction"]],
                            categorical=False)

    # Return the full tree in tree.py format.
    return get_tree(1)


def build_tree(conf, cp, input_file):
    """
    Returns a tree.py CART built from the input file data

    conf: configuration options
    cp: complexity parameter
    input_file: contains the data on which we apply CART regression, the last
    column is the response, the starting columns are the factors.
    """

    temp_dir = tempfile.mkdtemp()
    ti = os.path.join(temp_dir, "rcart.R")
    to = os.path.join(temp_dir, "out")
    categories = []
    rfile = open(ti, "w")

    # Prepare an R script that applies CART to the input data
    rfile.write('library(rpart)\n')
    rfile.write('data = read.table("{0}")\n'.
            format(input_file))

    for i, f in enumerate(conf["factors"]):
        if f["type"] == "categorical":
            rfile.write('data$V{0} = as.factor(data$V{1})\n'
                        .format(i + 1, i + 1))
            categories.append(f["values"])
        else:
            categories.append(f["range"])

    rfile.write("rtree = rpart(data$V{0} ~ . , data=data, "
                "control=rpart.control(cp={1}))\n".format(i + 2, cp))

    # Try to do cross validation pruning if enough samples are available
    rfile.write('if ("xerror" %in% colnames(rtree$cptable)) {')
    rfile.write(
            'bcp=rtree$cptable[which.min(rtree$cptable[,"xerror"]),"CP"]\n')
    rfile.write('rtree = prune(rtree, bcp)\n}\n')
    rfile.write("print(rtree)\n")
    rfile.close()

    # Launch R to build the CART tree
    print "Building CART model ..."
    os.system("R --slave --no-save < {0} > {1}".format(ti, to))

    # Parse and return the CART tree
    tf = open(to, "r")
    print "Parsing CART tree"
    T = parse_cart_tree(tf.readlines(), categories)
    tf.close()
    return T



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from test_util import CommandLineT


class FactorTests(CommandLineT):
    def test_ranges(self):
        """ Ranges should make sense """
        self.conf({"factors": [
          {"name":"f1",
           "type":"integer",
           "range":{"min":100, "max":10}}],
          "output_directory": "outdir",
          "modules": {
            "bootstrap": {"executable": "/bin/false"},
            "sampler": {"executable": "/bin/false"},
            "source": {"executable": "/bin/false"},
            "control": {"executable": "/bin/false"}}})

        r = self.run("ask test.conf", expect_error=True)
        assert r.returncode != 0, "Return code should be != 0"
        assert r.stderr.find("min") != -1,\
                 "error message should mention problem with min"

    def test_invalid_types(self):
        """  Check that unknown factor type is rejected """
        self.conf({"factors": [
          {"name":"f1",
           "type":"notvalidtype"}],
          "output_directory": "outdir",
          "modules": {
            "bootstrap": {"executable": "/bin/false"},
            "sampler": {"executable": "/bin/false"},
            "source": {"executable": "/bin/false"},
            "control": {"executable": "/bin/false"}}})

        r = self.run("ask test.conf", expect_error=True)
        assert r.returncode != 0, "Return code should be != 0"
        assert r.stderr.find("type") != -1,\
                 "error message should mention problem with type"

    def test_bad_categorical(self):
        """ Categorical factor missing values """
        self.conf({"factors": [
          {"name":"f1",
           "type":"categorical",
           "range": {"min": 0, "max": 10}}],
          "output_directory": "outdir",
          "modules": {
            "bootstrap": {"executable": "/bin/false"},
            "sampler": {"executable": "/bin/false"},
            "source": {"executable": "/bin/false"},
            "control": {"executable": "/bin/false"}}})

        r = self.run("ask test.conf", expect_error=True)
        assert r.returncode != 0, "Return code should be != 0"
        assert r.stderr.find("Values") != -1,\
                 "error message should mention problem with values"

    def test_bad_integer(self):
        """ Integer factor missing range """
        self.conf({"factors": [
          {"name":"f1",
           "type": "integer"}],
          "output_directory": "outdir",
          "modules": {
            "bootstrap": {"executable": "/bin/false"},
            "sampler": {"executable": "/bin/false"},
            "source": {"executable": "/bin/false"},
            "control": {"executable": "/bin/false"}}})

        r = self.run("ask test.conf", expect_error=True)
        assert r.returncode != 0, "Return code should be != 0"
        assert r.stderr.find("range") != -1,\
                 "error message should mention problem with range"

    def test_repeated_name(self):
        """ No repeated name in factors """
        self.conf({"factors": [
          {"name":"f1",
           "type":"integer",
           "range": {"min": 0, "max": 10}},
          {"name":"f1",
           "type":"integer",
           "range": {"min": 0, "max": 10}}],
          "output_directory": "outdir",
          "modules": {
            "bootstrap": {"executable": "/bin/false"},
            "sampler": {"executable": "/bin/false"},
            "source": {"executable": "/bin/false"},
            "control": {"executable": "/bin/false"}}})

        r = self.run("ask test.conf", expect_error=True)
        assert r.returncode != 0, "Return code should be != 0"
        assert r.stderr.find("name") != -1,\
                 "error message should mention problem with name"

    def test_types_ok(self):
        """ Check that valid factors are accepted """

        self.writefile("test.data", "0 5.5 0")

        self.conf({"factors": [
          {"name":"f1",
           "type":"integer",
           "range": {"min": 0, "max": 10}},
          {"name":"f2",
           "type":"float",
           "range": {"min": 0.5, "max": 10.33}},
          {"name":"f3",
           "type":"categorical",
           "values": ["a", "b", "foo"]}],
          "output_directory": "outdir",
          "modules": {
            "bootstrap": {"executable": "bootstrap/random-file",
                          "params": {"n": 1, "data_file": "test.data"}},
            "sampler": {"executable": "/bin/false"},
            "source": {"executable": "source/file",
                       "params": {"data_file": "test.data"}},
            "control": {"executable": "control/points",
                        "params": {"n": 0}}}})

        self.run("ask test.conf", expect_error=False)



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from test_util import CommandLineT


class BootstrapRandomFileTests(CommandLineT):
    def test_random_missing_params(self):
        """ Random-File bootstrap should print a message if parameters are missing
        """
        self.conf({
            "factors": [
            {"name":"f1", "type":"integer", "range":{"min":1, "max":8}},
            {"name":"f2", "type":"integer", "range":{"min":5, "max":4999}}],
            "modules": {"bootstrap":
                        {"params": {"data_file": self.tfile("test1.data")}}}})

        r = self.run_module("bootstrap/random-file",
                            "test.conf sample",
                            expect_error=True)
        assert "n" in r.stderr, "error message should mention n"

    def test_wrong_parameter(self):
        """ Random-File bootstrap should fail nicely when n is not integer """
        self.conf({
             "factors": [
            {"name":"f1", "type":"integer", "range":{"min":1, "max":8}},
            {"name":"f2", "type":"integer", "range":{"min":5, "max":4999}}],
             "modules": {"bootstrap":
                         {"params": {"n": "foo", "data_file": "bar"}}}})

        r = self.run_module("bootstrap/random-file",
                            "test.conf sample",
                            expect_error=True)
        assert r.stderr.find("int") != -1,\
                 "error message should mention expected type"
        assert "n" in r.stderr, "error message should mention n"

    def test_sampling(self):
        """ Random-File bootstrap should sample points """
        card = 17
        self.conf({
             "factors": [
            {"name":"f1", "type":"integer", "range":{"min":1, "max":8}},
            {"name":"f2", "type":"integer", "range":{"min":5, "max":4999}}],
             "modules": {"bootstrap":
                         {"params": {"n": card,
                                     "data_file": self.tfile("test1.data")}}}})

        r = self.run_module("bootstrap/random-file", "test.conf sample")
        assert "sample" in r.files_created, "sample is created"
        sample = r.files_created["sample"]
        sample_card = len([l for l in sample.bytes.split("\n") if l])
        assert sample_card == card, "we should get exactly 17 samples"

    def test_too_large_sampling(self):
        """ Asking more points than available in the file should fail """
        data = self.tfile("test1.data")
        card = len(file(data).readlines()) + 1
        self.conf(
            {"modules": {"bootstrap":
                         {"params": {"n": card, "data_file": data}}}})

        r = self.run_module("bootstrap/random-file",
                            "test.conf sample",
                            expect_error=True)
        assert r.stderr.find("Not enough points") != -1,\
                 "error message should mention expected type"


class BootstrapRandomTests(CommandLineT):
    def test_random_with_types(self):
        """ Test Random bootstrap"""

        card = 25
        self.writefile("test.data", "0 5.5 0")

        self.conf({"factors": [
          {"name":"f1",
           "type":"integer",
           "range": {"min": 0, "max": 10}},
          {"name":"f2",
           "type":"float",
           "range": {"min": 0.5, "max": 10.33}},
          {"name":"f3",
           "type":"categorical",
           "values": ["a", "b", "foo"]}],
          "output_directory": "outdir",
          "modules": {
            "bootstrap": {"executable": "bootstrap/random",
                          "params": {"n": 25}},
            "sampler": {"executable": "/bin/false"},
            "source": {"executable": "source/file",
                       "params": {"data_file": "test.data"}},
            "control": {"executable": "control/points",
                        "params": {"n": 0}}}})

        r = self.run("ask test.conf", expect_error=False)
        assert "outdir/requested00000.data" in r.files_created, "sample is created"
        sample = r.files_created["outdir/requested00000.data"]
        sample_card = len([l for l in sample.bytes.split("\n") if l])
        assert sample_card == card, "we should get exactly 25 samples"


class BootstrapLatinsquareTests(CommandLineT):
    def test_latinsquare_missing_params(self):
        """ Latinsquare bootstrap should print a message
        if parameters are missing
        """
        self.conf(
            {"factors": [],
            "modules": {"bootstrap":
                        {"params": {}}}})

        r = self.run_module("bootstrap/latinsquare",
                            "test.conf sample",
                            expect_error=True)
        assert "n" in r.stderr, "error message should mention n"

    def test_wrong_parameter(self):
        """ Latinsquare bootstrap should fail nicely when n is not integer """
        self.conf(
            {"factors": [],
             "modules": {"bootstrap":
                         {"params": {"n": "foo"}}}})

        r = self.run_module("bootstrap/latinsquare",
                            "test.conf sample",
                            expect_error=True)
        assert r.stderr.find("int") != -1,\
                 "error message should mention expected type"
        assert "n" in r.stderr, "error message should mention n"

    def test_sampling(self):
        """ Latinsquare bootstrap should sample points """
        card = 17
        self.run("mkdir outdir")
        self.conf(
            {"output_directory" : "outdir",
             "factors": [{"name": "f1", "type": "integer",
                          "range": {"min": -10, "max": 100}},
                         {"name": "f2", "type": "float",
                          "range": {"min": -10, "max": 100}},
                         {"name": "f3", "type": "categorical",
                          "values": ["a", "b", "c"]}],
             "modules": {"bootstrap":
                         {"params": {"n": card}}}})

        r = self.run_module("bootstrap/latinsquare", "test.conf sample")
        assert "sample" in r.files_created, "sample is created"
        sample = r.files_created["sample"]
        sample_card = len([l for l in sample.bytes.split("\n") if l])
        assert sample_card == card, "we should get exactly 17 samples"

class BootstrapLowDiscrepancyTests(CommandLineT):
    def test_lowdiscrepancy_missing_params(self):
        """ Lowdiscrepancy bootstrap should print a message
        if parameters are missing
        """
        self.conf(
            {"factors": [],
            "modules": {"bootstrap":
                        {"params": {}}}})

        r = self.run_module("bootstrap/lowdiscrepancy",
                            "test.conf sample",
                            expect_error=True)
        assert "n" in r.stderr, "error message should mention n"

    def test_wrong_parameter(self):
        """ Lowdiscrepancy bootstrap should fail nicely when n is not integer 
        """
        self.conf(
            {"factors": [],
             "modules": {"bootstrap":
                         {"params": {"n": "foo"}}}})

        r = self.run_module("bootstrap/lowdiscrepancy",
                            "test.conf sample",
                            expect_error=True)
        assert r.stderr.find("int") != -1,\
                 "error message should mention expected type"
        assert "n" in r.stderr, "error message should mention n"

    def test_sampling(self):
        """ Lowdiscrepancy bootstrap should sample points """
        card = 17
        self.run("mkdir outdir")
        self.conf(
            {"output_directory" : "outdir",
             "factors": [{"name": "f1", "type": "integer",
                          "range": {"min": -10, "max": 100}},
                         {"name": "f2", "type": "float",
                          "range": {"min": -10, "max": 100}},
                         {"name": "f3", "type": "categorical",
                          "values": ["a", "b", "c"]}],
             "modules": {"bootstrap":
                         {"params": {"n": card}}}})

        r = self.run_module("bootstrap/lowdiscrepancy", "test.conf sample")
        assert "sample" in r.files_created, "sample is created"
        sample = r.files_created["sample"]
        sample_card = len([l for l in sample.bytes.split("\n") if l])
        assert sample_card == card, "we should get exactly 17 samples"



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

def linear(reg,*args):
    """
    reg: a numpy vector that contains the coefficients of a n-dim affine equation.
         (the constant intercept is last).
    args: the n coordinates of a point.

    returns the value of the affine equation for the passed point.

    For example to compute f(x) = 2.x + 1 when x = 2,
   we can call:
    >>> import numpy as np
    >>> linear(np.array([ 2.,  1.]), 2)
    5.0
    """
    f = reg[-1]
    assert(len(reg)-1 == len(args))
    for p,v in zip(reg, args):
        f+=p*v
    return f

if __name__ == "__main__":
    import doctest
    doctest.testmod()



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from test_util import CommandLineT

class DriverTests(CommandLineT):
    def test_no_arguments(self):
        """ ask should complain if called without arguments """
        r = self.run("ask", expect_error=True)

        assert "usage" in r.stderr, "Usage message should be printed"
        assert r.returncode != 0, "Return code should be != 0"

    def test_no_output_rewrite(self):
        """ The output directory should not be overwriten """
        self.conf({"output_directory": "outdir",
                   "factors" : [],
                   "modules" : {
                      "bootstrap" : {"executable" : "/bin/false"},
                      "sampler" : {"executable" : "/bin/false"},
                      "source" : {"executable" : "/bin/false"},
                      "control" : {"executable" : "/bin/false"},
                   }})

        self.run("mkdir outdir")
        self.run("touch outdir/MARKER")
        r = self.run("ask test.conf", expect_error=True)
        assert r.returncode != 0, "Return code should be != 0"
        assert "outdir/MARKER" in r.files_after, "outdir should not be overwritten"

    def test_output_rewrite(self):
        """ The output directory should be overwriten when
            passing the --force_overwrite flag
        """
        self.conf({"output_directory": "outdir",
                   "factors" : [],
                   "modules" : {
                      "bootstrap" : {"executable" : "/bin/false"},
                      "sampler" : {"executable" : "/bin/false"},
                      "source" : {"executable" : "/bin/false"},
                      "control" : {"executable" : "/bin/false"},
                   }})

        self.run("mkdir outdir")
        self.run("touch outdir/MARKER")
        r = self.run("ask test.conf --force_overwrite", expect_error=True)
        assert r.returncode != 0, "Return code should be != 0"
        assert "outdir/MARKER" not in r.files_after, "outdir should be overwritten"



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from test_util import CommandLineT


class ControlTests(CommandLineT):
    def test_points_control_stop(self):
        """ If enough points reached control should stop """
        self.conf({
            "output_directory": "outdir",
            "modules": {
            "control": {
            "executable": "control/points",
            "params": {"n": 150}}}})

        data = self.tfile("test1.data")
        r = self.run_module("control/points",
                            "test.conf 1 {0} fakemodel".format(data),
                            expect_error=True)
        assert r.returncode == 254, "control should stop with 254 return code"

    def test_points_control_continue(self):
        """ If not enough points reached control should continue """
        self.conf({
            "output_directory": "outdir",
            "modules": {
            "control": {
            "executable": "control/points",
            "params": {"n": 2 ** 32}}}})

        data = self.tfile("test1.data")
        r = self.run_module("control/points",
                            "test.conf 1 {0} fakemodel".format(data),
                            expect_error=True)
        assert r.returncode == 0, "control should continue with 0 return code"



# Copyright (c) 2011-2012, Universite de Versailles St-Quentin-en-Yvelines
#
# This file is part of ASK.  ASK is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import json
import math
import os
import os.path
import shutil

from common.util import fatal

 # Modules that need to be defined for ask to work
expected_modules = ["sampler", "reporter", "control",
                 "model", "bootstrap", "source"]


def check_executable(modulepath):
    # Find exact module path
    # try raw module path
    if not os.path.isfile(modulepath):
        # if not found look relatively to ASKHOME
        modulepath = os.path.join(os.environ["ASKHOME"], modulepath)
        if not os.path.isfile(modulepath):
            fatal("Module {0} could not be found".format(modulepath))

    # Check if the found path is executable
    if not os.access(modulepath, os.X_OK):
        fatal("Module {0} is not executable".format(modulepath))

    return modulepath


class Configuration():
    def __init__(self,
                 user_configuration,
                 reload_live_configuration=True,
                 force_overwrite=False,
                 replay_only=False):

        # If reloading a live configuration, there is
        # no need to check it for errors nor load the
        # default one.
        if reload_live_configuration:
            self._conf = Configuration.load(user_configuration)
        else:
            # First load the default configuration
            default_configuration = os.path.join(os.environ["ASKHOME"],
                                                 "default.conf")
            self._conf = Configuration.load(default_configuration)

            # Now load the user configuration
            user_conf = Configuration.load(user_configuration)

            # Overwrite default configuration with user's one
            Configuration.overwrite(self._conf, user_conf)

            # Check and eventually create the output directory
            self.check_outdir(force_overwrite, replay_only)

            # Change and check modules paths
            self.check_modules()

            # Check the factors
            self.check_factors()

            # Write live configuration to disk
            filename = os.path.join(self._conf["output_directory"],
                                    ".ask.conf")
            self._conf["configuration_file"] = filename
            f = open(filename, "w")
            json.dump(self._conf, f, indent=2)
            f.write("\n")
            f.close()

    def check_outdir(self, force_overwrite, replay_only):
        """ Ensure that the output directory can be used """

        if os.path.exists(self("output_directory")):
            if force_overwrite:
                print("Removing existing output directory {0}"
                      " since --force_overwrite was passed."
                      .format(self("output_directory")))
                shutil.rmtree(self("output_directory"))
                os.makedirs(self("output_directory"))
            elif not replay_only:
                fatal("Output directory \"{0}\" already exists."
                      " Use --force_overwrite option to discard it."
                      .format(self("output_directory")))
        elif replay_only:
                fatal("Output directory \"{0}\" does not exist."
                      " Cannot use --replay_only."
                      .format(self("output_directory")))
        else:
            os.makedirs(self("output_directory"))

    def check_factors(self):
        """ Check the factors configuration section """

        if "factors" not in self._conf:
            fatal("Configuration file is missing the factors section")
        if not isinstance(self._conf["factors"], list):
            fatal("Factors section should contain a list of factors")

        names = set()
        for f in self._conf["factors"]:
            if "name" not in f:
                fatal("Some factors are missing a name")
            if f["name"] in names:
                fatal("Two factors have the same name: {0}"
                      .format(f["name"]))
            else:
                names.add(f["name"])

            if "type" not in f:
                fatal("Factor {0} is missing a type"
                      .format(f["name"]))

            if f["type"] == "integer":
                if "range" not in f or "min" not in f["range"] \
                   or "max" not in f["range"]:
                    fatal("Factor {0} needs a valid range")

                # Check that min and max are integers and
                # that min < max
                mi = f["range"]["min"]
                ma = f["range"]["max"]
                if math.modf(mi)[0] != 0 or math.modf(ma)[0] != 0:
                    fatal("Range bounds for factor {0} should be integers"
                          .format(f["name"]))

                if mi >= ma:
                    fatal("Range of factor {0} is invalid : min >= max"
                          .format(f["name"]))

            elif f["type"] == "float":
                if "range" not in f or "min" not in f["range"] \
                   or "max" not in f["range"]:
                    fatal("Factor {0} needs a valid range")

                # Check that min < max
                if f["range"]["min"] >= f["range"]["max"]:
                    fatal("Range of factor {0} is invalid : min >= max"
                          .format(f["name"]))

            elif f["type"] == "categorical":
                if "values" not in f or not isinstance(f["values"], list):
                    fatal("Values field is missing or invalid for factor {0}"
                          .format(f["name"]))
            else:
                fatal("Unknown type {0} for factor {1})"
                      .format(f["type"], f["name"]))

    def check_modules(self):
        """ Check the modules configuration section
            and replace module relative paths with absolute ones
        """
        for module in expected_modules:
            # Check that the module executable can be found,
            # and update it with its complete path
            modulepath = self("modules.{0}.executable"
                           .format(module))
            self._conf["modules"][module]["executable"] =\
              check_executable(modulepath)

            if module == "model":
                # Model module as an optional executable, predictor
                # which we may need to check
                if "predictor" in self._conf["modules"][module]:
                    predictorpath = self._conf["modules"][module]["predictor"]
                    self._conf["modules"][module]["predictor"] =\
                      check_executable(predictorpath)

    @staticmethod
    def load(filename):
        # Parse configuration
        try:
            f = open(filename, "r")
            conf = json.load(f)
            f.close()
        except ValueError, e:
            fatal("Problem parsing configuration file {0}:\n"
                  "  {1}".format(filename, e))
        except IOError:
            fatal("Could not open configuration file {0}"
                  .format(filename))
        return conf

    @staticmethod
    def overwrite(old_conf, new_conf):
        """
        Overwrite a previous configuration with a new one.  Existing
        configuration values not redefined by new_conf are preserved
        (which allows to merge different configurations together).

        The merged configuration is written to old_conf.
        """

        for key, value in new_conf.iteritems():
            if isinstance(value, dict) and key in old_conf:
                Configuration.overwrite(old_conf[key],
                                        value)
            else:
                old_conf[key] = value

    def __call__(self, key, expected_type=None, default_value=None):
        subkeys = key.split(".")
        # Check that the key exists and handle default_value
        try:
            V = self._conf
            for sk in subkeys:
                V = V[sk]
        except KeyError:
            if default_value == None:
                fatal("Missing configuration parameter {0}"
                      .format(key))
            else:
                V = default_value

        # Check the type
        if expected_type and not isinstance(V, expected_type):
            fatal("Wrong parameter {0} : expected {1} got '{2}'"
                    .format(key, expected_type.__name__, V))
        else:
            return V

    def __getitem__(self, key):
        return self._conf[key]
